#!/usr/bin/env python
# -*- coding: utf-8 -*-

############################################################################
# casetta.py - Rosetta Stone for program files for casio computers
# (c) 2006 Florian Birée aka Thesa <florian.biree@laposte.net> ~ http://filyb.info/
#
# Version 0.1.2
#
# changelog:
#
# # 0.1.2 version:
#
#   o 'or' bug fix in the fxiDecode function
#   o change the way to found the name of a data in the fxi format (work with any name length)
# 
# # 0.1.1 version:
#
#   o many bug fix
#
# # 0.1.0 version:
#
#   o first release
#
############################################################################
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# http://www.gnu.org/copyleft/gpl.html
#
############################################################################

__author__ = "Florian Birée"
__version__ = "0.1.2"
copyright = "Florian Birée, 2006, under GNU/GPL (http://www.gnu.org/copyleft/gpl.html)"

__doc__ = """
casetta.py - python module for convertions into casio programs files
You can use this module in command line:
Usage: casetta [INPUT] [INPUT_FORMAT] [OUTPUT] [OUTPUT_FORMAT]

  INPUT and OUTPUT must be a filename, a path, or the keyword 'casio', depend of the format.
  If format isn't specified, caetta will try to find it thanks to the filename extension.
  This may fail in some case, just specifie the format.
  
  Avaible formats are:
  
  casio     receive or send the data to the calculator. Optional (the keyword casio is enough).
  cafix     original cafix format. For output, files will be saved in the dirname of OUTPUT.
  cat       Cat format, used by the program Fa-123
  fxi       Fxi format, used by the program Fx-Interface (only usable in input!)
  newcat    New Cat format, a cat-like format with more equivalents of the special characters, and
            unix returns to the line (to edit yours programs)

  Warning: only programs data can be converted! Others data (picture, matrix...) may be converted
  in some formats (cafix), but nothing is sure.

"""
__doc__ += "Version %s by %s\n"%(__version__,__author__)
__doc__ += "Copyright (C), %s"%copyright

#Import catdic
import os, sys
from catdic import *

#data Type constants
none = 0
program = 1
unknown = -1

#file format constants
unknown = -1
casio = -2
cafix = 0
cat = 1
fxi = 2
newcat = 3 #Cat format + alias for others characters (for programers) + unix line return

#Build the raw2fxi dic
#Thanks to http://www.casioland.net/tutoriaux.php?idTuto=3&noHeader=1
r2f = {}
dep, a, i, j = 23, 255, 0, 0
v = dep
while a >= 0 :
    v = v + 1
    r2f[a] = v
    a -= 1
    i += 1
    if i == 8 :
        dep -= 8
        j += 1
        i = 0
        v = dep
    if j == 4 :
        j = 0
        dep += 8 * 8
        v = dep

#Build the fxi2raw dic
f2r = {}
for a, f in r2f.items():
    f2r[f] = a


#Data manipulation classes:
class Data:

    def __init__(self):
        """ Make an empty data (type=0) """
        self.name = ''
        self.rawData = ''
        self.dType = 0
   
    def getText(self):
        """ Convert rawData in cat-like text (only for programs) """
        if self.dType == program :
            return textEncode(self.rawData)
        
    def setText(self, text):
        """ Convert cat-like text in rawData (only for programs) """
        if self.dType == program :
            self.rawData = textDecode(text)
        
        
class FileData:

    def __init__(self):
        """ Make an empty FileData """
        self.data = []
        self.export = []

    def delData(self,i):
        """ Del the data with the index i """
        del(self.data[i])
        #clear self.export (index list has changed)
        self.export = []
    
    def newData(self):
        """ Make a new data, and return it index """
        self.data.append(Data())
        return len(self.data) - 1
    
    def save(self, filename, format = ''):
        """ save in (a) file(s) the data """
        if format == '' :
            format = chooseFormat(filename)
            if format == unknown :
                return
        
        if format == cafix :
            cafixSave(filename,self)
        elif format == cat :
            catSave(filename,self)
        elif format == fxi :
            fxiSave(filename,self)
        elif format == newcat :
            newcatSave(filename,self)
    
    def addData(self, fileData):
        """ add data from another fileData """
        o = len(self.data)
        self.data += fileData.data
        return range(o,len(self.data))
    
    def add(self, filename):
        """ open datas from filename and add them in the file 
        Return the list of new index
        """
        if filename[-4:].lower() == '.fxi' :
            return self.addData(fxiOpen(filename))
        elif filename[-4:].lower() == '.cat' :
            return self.addData(catOpen(filename))
        elif filename[-5:].lower() == 'cafix' :
            return self.addData(cafixOpen(filename))
        elif filename[-6:].lower() == 'newcat' :
            return self.addData(newcatOpen(filename))

#Decode/Encode functions

def fxiEncode(rawData):
    """ Convert raw data to fxi-like data """
    out = ''
    for c in range(len(rawData)) :
        out += chr(r2f[ord(rawData[c])])
    return out

def fxiDecode(fxiData):
    """ Convert fxi-like data to raw data """
    out = ''
    for c in range(len(fxiData)) :
        out += chr(f2r[ord(fxiData[c])])
    #Fix the 'or' bug (some 'or' strings was replaced by a 0xAA string ##v0.1.2
    out = out.replace(chr(0xAA),'or')
    return out

def catEncode(rawData):
    """ Convert raw data to cat-like data """
    out = rawData
    #convert * /
    out = out.replace("*", "\\aster")
    out = out.replace("/", "\\slash")
    #Change cat caract.
    for rawCar in r2c :
        out = out.replace(rawCar,r2c[rawCar])
    #Change last cat caract.
    for rawCar in r2cl :
        out = out.replace(rawCar,r2cl[rawCar])
    #Change octet-cat caract.
    for rawCar in or2c :
        out = out.replace(rawCar,or2c[rawCar])
    #Remove the end caract (chr(255))
    out = out.replace(chr(255),'')
    #Change end line (chr(13) to chr(13)+chr(10)
    out = out.replace(chr(13),chr(13)+chr(10))
    return out

def catDecode(catData):
    """ Convert cat-like data to raw data """
    out = catData
    #Change end line chr(13)+chr(10) in chr(13)
    out = out.replace(chr(13)+chr(10),chr(13))
    #Change cat caract.
    for catEquiv in c2r :
        out = out.replace(catEquiv,c2r[catEquiv])
    #Change octet-cat caract.
    for catEquiv in oc2r :
        out = out.replace(catEquiv,oc2r[catEquiv])
    #Change last cat caract.
    for catEquiv in c2rl :
        out = out.replace(catEquiv,c2rl[catEquiv])
    #convert * /
    out = out.replace("\\aster","*")
    out = out.replace("\\slash","/")
    #Add the last caract (chr(255))
    out += chr(255)
    return out

def textEncode(rawData):
    """ Convert raw data to Editable text (NewCat format) """
    out = rawData
    #convert * /
    out = out.replace("*", "\\aster")
    out = out.replace("/", "\\slash")
    #Change cat caract.
    for rawCar in r2c :
        out = out.replace(rawCar,r2c[rawCar])
    #Change new cat caract.
    for rawCar in r2n :
        out = out.replace(rawCar,r2n[rawCar])
    #Change octet-cat caract. *->\aster
    for rawCar in or2c :
        out = out.replace(rawCar,or2c[rawCar])
    #Change new octet-cat caract.
    for rawCar in or2n :
        out = out.replace(rawCar,or2n[rawCar])
    #Change last cat caract. (169)->*
    for rawCar in r2cl :
        out = out.replace(rawCar,r2cl[rawCar])
    #Remove the end caract (chr(255))
    out = out.replace(chr(255),'')
    #Change end line chr(13) in chr(10)
    out = out.replace(chr(13),'\n')
    return out
    
def textDecode(textData):
    """ Convert Editable text (NewCat format) to raw data """
    out = textData
    #Change end line chr(10) in chr(13)
    out = out.replace('\n',chr(13))
    #Change cat caract.
    for catEquiv in c2r :
        out = out.replace(catEquiv,c2r[catEquiv])
    #Change new cat caract.
    for catEquiv in n2r :
        out = out.replace(catEquiv,n2r[catEquiv])
    #Change new octet-cat caract.
    for catEquiv in on2r :
        out = out.replace(catEquiv,on2r[catEquiv])
    #Change octet-cat caract. \-> -> (14)
    for catEquiv in oc2r :
        out = out.replace(catEquiv,oc2r[catEquiv])
    #Change last cat caract. *->(169)
    for catEquiv in c2rl :
        out = out.replace(catEquiv,c2rl[catEquiv])
    #convert * /
    out = out.replace("\\aster","*")
    out = out.replace("\\slash","/")
    #Add the last caract (chr(255))
    out += chr(255)
    return out


# Files format manager
def chooseFormat(filename):
    """ Return the format of filename (with extension), -1 is not found """
    if filename[-4:].lower() == '.fxi' :
        return fxi
    elif filename[-4:].lower() == '.cat' :
        return cat
    elif filename[-6:].lower() == '.cafix' :
        return cafix
    elif filename[-7:].lower() == '.newcat' :
        return newcat
    else:
        return unknown

def fileOpen(filename):
    """ Run the *Open function corresponding with the filename """
    openFunc = {cafix:cafixOpen,cat:catOpen,fxi:fxiOpen,newcat:newcatOpen}
    return openFunc[chooseFormat(filename)](filename)

def fxiOpen(filename):
    """ Open a file in the fxi Format (2) """
    #Organization of a FXI file:
    ##Many noise ??
    #(First record:)
    ##TEXT?PG
    ##4*?
    ##Name
    ##0xFF * (16+8-len(name))
    ##NL
    ##12*0xFF
    ##prog...
    ##0xFF
    ##Many noise ??
    #(second tecord)
    
    f = open(filename,'r')
    #Build the FileData
    fd = FileData()
    #Get the raw data
    ctn = fxiDecode(f.read())
    f.close()
    #Delete noise (in the order !!):
    #form 24 to 16 FF in the data header (depend of the lenght of files names) ##v0.1.2
    ffLen = range(16,25)
    ffLen.reverse()
    for l in ffLen :
        ctn = ctn.replace(l*chr(0xFF),'')
    #12 FF in the data header
    ctn = ctn.replace(12*chr(0xFF),12*chr(0xFE)) #used to split the record into program header and program
    #two FF in the head of the file
    ctn = ctn.replace(2*chr(0xFF), '')
    #split data (by one FF character)
    for record in ctn.split(chr(0xFF)) :
        #record is now like that : (? is variable data)
        #ManyData (?)
        #TXT?PG????Name_of_the_prgNL[12*0xFE]programm
        #for each data, set metadata and raw data
        if 'NL' + 12*chr(0xFE) in record :
            cd = fd.newData()
            header = record.split('NL' + 12*chr(0xFE))[0]
            fd.data[cd].rawData = record.split('NL' + 12*chr(0xFE))[1] + chr(0xFF)
            fd.data[cd].name = header[header.find('TXT')+10:]
            fd.data[cd].dType = program #If not... "advienne que pourra" ;-)    
    return fd

def fxiSave(filename, fileData):
    """ Save a file in the fxi Format (2) """
    print "This functions is not implemented yet :-/"
    #For each data, convert in fxi format
    #Set the header and metadata
    #join in a fxi file

def catOpen(filename):
    """ Open a file in the cat Format (1) """
    #Build the FileData
    fd = FileData()
    f = open(filename, 'r')
    ctn = f.read()
    #Split data
    for record in ctn.split('%End' + chr(13)+chr(10)) :
        if '%Data Record' in record :
            header = record.split('%Data Record' + chr(13)+chr(10))[0]
            data = record.split('%Data Record' + chr(13)+chr(10))[1]
            #for each data, set metadata
            dt = unknown
            format = 'raw'
            name = 'unknown'
            for headline in header.split(chr(13)+chr(10)) :
                label = headline.split(':')[0]
                try :
                    value = headline.split(':')[1]
                except :
                    value = ''
                if label == 'Data Type' and value == 'PG' :
                    dt = program
                elif label == 'Format' and value == 'TXT' :
                    format = 'txt'
                elif label == 'File Name' :
                    name = value
            #Creat the data
            cd = fd.newData()
            fd.data[cd].name = name
            fd.data[cd].dType = dt
            if format == 'raw' :
                fd.data[cd].rawData = data
            elif format == 'txt' :
                fd.data[cd].rawData = catDecode(data)
    return fd

def catSave(filename, fileData):
    """ Save a file in the cat Format (1) """
    f = open(filename,'w')
    ctn = ''
    #For each data, convert in the cat format
    for data in fileData.data :
        dt = ''
        if data.dType == program :
            dt = "PG"
        if dt != '' :
            #Set the header and metadata
            ctn += "%Header Record"+chr(13)+chr(10)
            ctn += "Format:TXT"+chr(13)+chr(10)
            ctn += "Communication SW:0"+chr(13)+chr(10)
            ctn += "Data Type:" + dt + chr(13)+chr(10)
            ctn += "Capacity:" + str(len(data.rawData)) + chr(13)+chr(10)
            ctn += "File Name:" + data.name + chr(13)+chr(10)
            ctn += "Group Name:" + chr(13)+chr(10)
            ctn += "Password:" + chr(13)+chr(10)
            ctn += "Option1:NL" + chr(13)+chr(10)
            ctn += "Option2:" + chr(13)+chr(10)
            ctn += "Option3:" + chr(13)+chr(10)
            ctn += "Option4:" + chr(13)+chr(10)
            ctn += "%Data Record" + chr(13)+chr(10)
            #Write cat data
            ctn += catEncode(data.rawData)
            ctn += "%End" + chr(13)+chr(10)
        else :
            if data.dType != 0 :
                print "This data type (" + str(data.dType) + ") is not supported in write access for cat format"
    #join in a cat file
    f.write(ctn)
    f.close()

def newcatOpen(filename):
    """ Open a file in the newcat Format (3) """
    #Build the FileData
    fd = FileData()
    f = open(filename, 'r')
    ctn = f.read()
    #Split data
    for record in ctn.split('%End\n') :
        if '%Data Record' in record :
            header = record.split('%Data Record\n')[0]
            data = record.split('%Data Record\n')[1]
            #for each data, set metadata
            dt = unknown
            format = 'raw'
            name = 'unknown'
            for headline in header.split('\n') :
                label = headline.split(':')[0]
                try :
                    value = headline.split(':')[1]
                except :
                    value = ''
                if label == 'Data Type' and value == 'PG' :
                    dt = program
                elif label == 'Format' and value == 'TXT' :
                    format = 'txt'
                elif label == 'File Name' :
                    name = value
            #Creat the data
            cd = fd.newData()
            fd.data[cd].name = name
            fd.data[cd].dType = dt
            if format == 'raw' :
                fd.data[cd].rawData = data
            elif format == 'txt' :
                fd.data[cd].rawData = textDecode(data)
    return fd

def newcatSave(filename, fileData):
    """ Save a file in the newcat Format (4) """
    f = open(filename,'w')
    ctn = ''
    #For each data, convert in the newcat format
    for data in fileData.data :
        dt = ''
        if data.dType == program :
            dt = "PG"
        if dt != '' :
            #Set the header and metadata
            ctn += "%Header Record\n"
            ctn += "Format:TXT\n"
            ctn += "Communication SW:0\n"
            ctn += "Data Type:" + dt + '\n'
            ctn += "Capacity:" + str(len(data.rawData)) + '\n'
            ctn += "File Name:" + data.name + '\n'
            ctn += "Group Name:\n"
            ctn += "Password:\n"
            ctn += "Option1:NL\n"
            ctn += "Option2:\n"
            ctn += "Option3:\n"
            ctn += "Option4:\n"
            ctn += "%Data Record\n"
            #Write *New* cat data
            ctn += textEncode(data.rawData)
            ctn += "%End\n"
        else :
            if data.dType != 0 :
                print "This data type (" + str(data.dType) + ") is not supported in write access for new cat format"
    #join in a cat file
    f.write(ctn)
    f.close()

def cafixOpen(filename):
    """ Open a file in the cafix Format (0) """
    #Build the FileData
    fd = FileData()
    cd = fd.newData()
    #Try to import metadata from the file name
    fd.data[cd].name = os.path.basename(filename).replace('prog-','').replace('.cafix','')
    #Open raw data
    f = open(filename,'r')
    fd.data[cd].rawData = f.read()
    f.close()
    #Try to set the type:
    if os.path.basename(filename)[:5] == 'prog-' :
        fd.data[cd].dType = program
    else :
        fd.data[cd].dType = unknown
    return fd
    
def cafixSave(directory, fileData):
    """ Save a file in the cafix Format (0) """
    folder = os.path.dirname(directory)
    #For each data
    for data in fileData.data :
        #make a filename from the metadata
        if len(fileData.data) == 1 and os.path.basename(directory) != '':
            #try to use directory as name
            filename = directory
        else :
            if data.dType == program :
                filename = os.path.join(folder,'prog-' + data.name + '.cafix')
            elif data.dType == unknown :
                filename = os.path.join(folder,'unknown-' + data.name + '.cafix')
        #write raw data
        f = open(filename,'w')
        f.write(data.rawData)
        f.close()

#Casio manager

def casioImport(directory):
    """ import all cafix files from directory in fileData """
    fd = FileData()
    for file in [file for file in os.listdir(directory) if os.path.isfile(os.path.join(directory,file)) and file[-6:] == '.cafix'] :
        fd.add(os.path.join(directory,file))
    return fd

def casioExport(directory, fileData):
    """ export all data from fileData wich have its index in fileData.export list in cafix format in directory """
    for i in fileData.export :
        f = open(os.path.join(directory,fileData.data[i].name),'w')
        f.write(fileData.data[i].rawData)
        f.close()

#Command line interface
if __name__ == '__main__' :
    io = ['input',unknown,'output',unknown]
    avaibleTypes = {'casio':casio,'cafix':cafix,'cat':cat,'fxi':fxi,'newcat':newcat}
    openFunc = {cafix:cafixOpen,cat:catOpen,fxi:fxiOpen,newcat:newcatOpen}
    saveFunc = {cafix:cafixSave,cat:catSave,fxi:fxiSave,newcat:newcatSave}
    #Get arguments
    if len(sys.argv) < 3 :
        print __doc__
        sys.exit()
    elif len(sys.argv) == 3 :
        #1:input,2:output
        io[0] = sys.argv[1]
        io[2] = sys.argv[2]
    elif len(sys.argv) == 4 :
        #1:input, 2:input_type|output, 3:output|output_type
        io[0] = sys.argv[1]
        if sys.argv[2] in avaibleTypes.keys() :
            io[1] = avaibleTypes[sys.argv[2]]
            io[2] = sys.argv[3]
        else :
            io[2] = sys.argv[2]
            io[3] = avaibleTypes[sys.argv[3]]
    else :
        #1:input, 2:input_type, 3:output, 3:output_type
        io = [sys.argv[1],avaibleTypes[sys.argv[2]],sys.argv[3],avaibleTypes[sys.argv[4]]]
    #Found loosing formats
    if io[1] == unknown and io[0] == 'casio' :
        io[1] = casio
    elif io[1] == unknown :
        io[1] = chooseFormat(io[0])
        if io[1] == unknown :
            print "casetta can't find the input format. Please specify it."
            sys.exit()
    if io[3] == unknown and io[2] == 'casio' :
        io[3] = casio
    elif io[3] == unknown :
        io[3] = chooseFormat(io[2])
        if io[3] == unknown :
            print "casetta can't find the output format. Please specify it."
            sys.exit()
    #####INPUT
    if io[1] != casio :
        #Creat the FileData from input file
        fd = openFunc[io[1]](io[0])
    else :
        #Import data from the calculator
        #Check the cafix command
        if len([path for path in os.defpath.split(os.pathsep) if os.path.exists(os.path.join(path,'cafix'))]) == 0 :
            print "The command cafix was not found in your path."
            print "While cafix is not in your path, you can't use the 'casio' keyword"
            print "Abort"
            sys.exit()
        #Check /dev/casio
        if not os.path.exists('/dev/casio') :
            print "Warning! the calculator port is not configured yet."
            print "You must run 'cafix-setup' as root"
            print "Press CTRL-C to quit this program and do it yourself"
            print "Press enter to run the command 'sudo cafix-setup'"
            raw_input()
            os.system('sudo cafix-setup')
            if not os.path.exists('/dev/casio') :
                print "/dev/casio not found. Abort"
                sys.exit()
            else :
                print "/dev/casio found."
        #Choice a temp directory and make it
        folder = os.tmpnam() + '-casetta'
        os.mkdir(folder)
        #Get the current dir
        curdir = os.path.abspath(os.path.curdir)
        #Run cafix -r in this dir
        print "You can send data from the calculator"
        os.system("cd " + folder + " && cafix -r")
        os.system("cd " + curdir)
        #import all cafix files
        fd = casioImport(folder)
    #####OUTPUT
    if io[3] != casio :
        #Save the FileData to output file
        fd.save(io[2],io[3])
    else :
        #Export the FileData to the calculator
        #Mark 'export' for all data in fd
        fd.export = [i for i in range(len(fd.data))]
        #Check cafix
        if len([path for path in os.defpath.split(os.pathsep) if os.path.exists(os.path.join(path,'cafix'))]) == 0 :
            print "The command cafix was not found in your path."
            print "While cafix is not in your path, you can't use the 'casio' keyword"
            print "Abort"
            sys.exit()
        #Check /dev/casio
        if not os.path.exists('/dev/casio') :
            print "Warning! the calculator port is not configured yet."
            print "You must run 'cafix-setup' as root"
            print "Press CTRL-C to quit this program and do it yourself"
            print "Press enter to run the command 'sudo cafix-setup'"
            raw_input()
            os.system('sudo cafix-setup')
            if not os.path.exists('/dev/casio') :
                print "/dev/casio not found. Abort"
                sys.exit()
            else :
                print "/dev/casio found."
        #Choice a temp directory and make it
        folder = os.tmpnam() + '-casetta'
        os.mkdir(folder)
        #Export data in this folder
        casioExport(folder,fd)
        #Get the current dir
        curdir = os.path.abspath(os.path.curdir)
        #Run cafix -s in this dir
        print "Run your calculator, and set it in the state 'Receiving'. Next press Enter."
        raw_input()
        os.system("cd " + folder + " && cafix -s *")
        os.system("cd " + curdir)
    #End
    print 'Done'
